﻿using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System;

namespace HeuristicLab.Problems.PTSP.MoveEvaluators.OneShift {
  class PTSPEstimatedInsertionEvaluator : PTSPPathMoveEvaluator, IPermutationTranslocationMoveOperator {

    public override Type EvaluatorType {
      get { return typeof(PTSPEstimatedInversionMovePathEvaluator); }
    }
    public ILookupParameter<TranslocationMove> TranslocationMoveParameter {
      get { return (ILookupParameter<TranslocationMove>)Parameters["TranslocationMove"]; }
    }

    public IValueParameter<ItemList<ItemList<IntValue>>> RealizationsParameter {
      get { return (IValueParameter<ItemList<ItemList<IntValue>>>)Parameters["Realizations"]; }
    }

    [StorableConstructor]
    protected PTSPEstimatedInsertionEvaluator(bool deserializing) : base(deserializing) { }
    protected PTSPEstimatedInsertionEvaluator(PTSPEstimatedInsertionEvaluator original, Cloner cloner) : base(original, cloner) { }
    public PTSPEstimatedInsertionEvaluator()
      : base() {
      Parameters.Add(new LookupParameter<TranslocationMove>("TranslocationMove", "The move to evaluate."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new PTSPEstimatedInsertionEvaluator(this, cloner);
    }

    public static double EvaluateByCoordinates(Permutation permutation, TranslocationMove move, DoubleMatrix coordinates, PTSPEstimatedInsertionEvaluator evaluator) {
      int edge1source = permutation.GetCircular(move.Index1 - 1);
      int edge1target = permutation[move.Index1];
      int edge2source = permutation[move.Index2];
      int edge2target = permutation.GetCircular(move.Index2 + 1);
      if (move.Index2 - move.Index1 >= permutation.Length - 2) return 0;
      double moveQuality = 0;
      // remove two edges
      moveQuality -= evaluator.CalculateDistance(coordinates[edge1source, 0], coordinates[edge1source, 1],
            coordinates[edge1target, 0], coordinates[edge1target, 1]);
      moveQuality -= evaluator.CalculateDistance(coordinates[edge2source, 0], coordinates[edge2source, 1],
        coordinates[edge2target, 0], coordinates[edge2target, 1]);
      // add two edges
      moveQuality += evaluator.CalculateDistance(coordinates[edge1source, 0], coordinates[edge1source, 1],
        coordinates[edge2source, 0], coordinates[edge2source, 1]);
      moveQuality += evaluator.CalculateDistance(coordinates[edge1target, 0], coordinates[edge1target, 1],
        coordinates[edge2target, 0], coordinates[edge2target, 1]);
      return moveQuality;
    }

    public static double EvaluateByDistanceMatrix(Permutation permutation, TranslocationMove move, DistanceMatrix distanceMatrix, ItemList<ItemList<IntValue>> realizations) {
      //permutation = new Permutation(PermutationTypes.Absolute, new int[]{0,1,2,3,4,5,6,7}); TODO remove test
      //move = new TranslocationMove(0, 0, 4); TODO remove test
      double moveQuality = 0;
      int[] edges = new int[12];
      int[] indices = new int[12];
      edges[0] = permutation.GetCircular(move.Index1 - 1);
      edges[6] = edges[0];
      indices[0] = move.Index1-1;
      if (indices[0] == -1) indices[0] = permutation.Length - 1;
      indices[6] = indices[0];
      edges[1] = permutation[move.Index1];
      indices[1] = move.Index1;
      edges[2] = edges[1];
      indices[2] = indices[1];
      edges[9] = edges[1];
      indices[9] = indices[1];
      edges[10] = edges[1];
      indices[10] = indices[1];
      edges[3] = permutation.GetCircular(move.Index1 + 1);
      edges[7] = edges[3];
      indices[3] = move.Index1+1;
      if (indices[3] == permutation.Length + 1) indices[3] = 0;
      indices[7] = indices[3];
      edges[4] = permutation[move.Index3];
      indices[4] = move.Index3;
      edges[8] = edges[4];
      indices[8] = indices[4];
      edges[5] = permutation.GetCircular(move.Index3 + 1);
      edges[11] = edges[5];
      indices[5] = move.Index3+1;
      if (indices[5] == permutation.Length + 1) indices[5] = 0;
      indices[11] = indices[5];
      int[] aPosteriori = new int[12];
      foreach (ItemList<IntValue> realization in realizations) {
        //int[] realization2 = new int[] { 1, 0, 1, 1, 0, 0, 1, 1 }; TODO remove test
        for (int i = 0; i < edges.Length; i++) {
          if (realization[edges[i]].Value == 1) {
            aPosteriori[i] = edges[i];
          } else {
            int j=1;
            if (i%2==0) {
              // find nearest predecessor in realization if source edge
              while(realization[permutation.GetCircular(indices[i]-j)].Value == 0) {
                j--;
              }
              aPosteriori[i] = permutation.GetCircular(indices[i] - j);
            } else {
              // find nearest successor in realization if target edge
              while(realization[permutation.GetCircular(indices[i]+j)].Value == 0) {
                j++;
              }
              aPosteriori[i] = permutation.GetCircular(indices[i] + j);
            }
          }
        }
        // compute cost difference between the two a posteriori solutions
        moveQuality += distanceMatrix[aPosteriori[6], aPosteriori[7]] + distanceMatrix[aPosteriori[8], aPosteriori[9]] + distanceMatrix[aPosteriori[10], aPosteriori[11]];
        moveQuality -= distanceMatrix[aPosteriori[0], aPosteriori[1]] - distanceMatrix[aPosteriori[2], aPosteriori[3]] - distanceMatrix[aPosteriori[4], aPosteriori[5]];
      }
      // return average of cost differences
      return moveQuality / realizations.Capacity;
    }

    protected override double EvaluateByCoordinates(Permutation permutation, DoubleMatrix coordinates) {
      return EvaluateByCoordinates(permutation, TranslocationMoveParameter.ActualValue, coordinates, this);
    }

    protected override double EvaluateByDistanceMatrix(Permutation permutation, DistanceMatrix distanceMatrix) {
      return EvaluateByDistanceMatrix(permutation, TranslocationMoveParameter.ActualValue, distanceMatrix, RealizationsParameter.Value);
    }

    protected override double CalculateDistance(double x1, double y1, double x2, double y2) {
      return Math.Round(Math.Sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)));
    }
  }
}
